@@ -164,17 +164,17 @@
   */
 static const char server_session_id_context[] = "Postfix/TLS";
 
-#ifndef OPENSSL_NO_TLSEXT
- /*
-  * We retain the cipher handle for lifetime of the process, it is not freed.
-  */
-static const EVP_CIPHER *tkt_cipher;
-#endif
-
 #define GET_SID(s, v, lptr)	((v) = SSL_SESSION_get_id((s), (lptr)))
 
+ /* OpenSSL 1.1.0 bitrot */
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
 typedef const unsigned char *session_id_t;
 
+#else
+typedef unsigned char *session_id_t;
+
+#endif
+
 /* get_server_session_cb - callback to retrieve session from server cache */
 
 static SSL_SESSION *get_server_session_cb(SSL *ssl, session_id_t session_id,
@@ -299,17 +299,19 @@ static int new_server_session_cb(SSL *ssl, SSL_SESSION *session)
 
 /* ticket_cb - configure tls session ticket encrypt/decrypt context */
 
-#if !defined(OPENSSL_NO_TLSEXT)
+#if defined(SSL_OP_NO_TICKET) && !defined(OPENSSL_NO_TLSEXT)
 
 static int ticket_cb(SSL *con, unsigned char name[], unsigned char iv[],
 		          EVP_CIPHER_CTX * ctx, HMAC_CTX * hctx, int create)
 {
     static const EVP_MD *sha256;
+    static const EVP_CIPHER *ciph;
     TLS_TICKET_KEY *key;
     TLS_SESS_STATE *TLScontext = SSL_get_ex_data(con, TLScontext_index);
     int     timeout = ((int) SSL_CTX_get_timeout(SSL_get_SSL_CTX(con))) / 2;
 
     if ((!sha256 && (sha256 = EVP_sha256()) == 0)
+	|| (!ciph && (ciph = EVP_get_cipherbyname(var_tls_tkt_cipher)) == 0)
 	|| (key = tls_mgr_key(create ? 0 : name, timeout)) == 0
 	|| (create && RAND_bytes(iv, TLS_TICKET_IVLEN) <= 0))
 	return (create ? TLS_TKT_NOKEYS : TLS_TKT_STALE);
@@ -317,13 +319,13 @@ static int ticket_cb(SSL *con, unsigned char name[], unsigned char iv[],
     HMAC_Init_ex(hctx, key->hmac, TLS_TICKET_MACLEN, sha256, NOENGINE);
 
     if (create) {
-	EVP_EncryptInit_ex(ctx, tkt_cipher, NOENGINE, key->bits, iv);
+	EVP_EncryptInit_ex(ctx, ciph, NOENGINE, key->bits, iv);
 	memcpy((void *) name, (void *) key->name, TLS_TICKET_NAMELEN);
 	if (TLScontext->log_mask & TLS_LOG_CACHE)
 	    msg_info("%s: Issuing session ticket, key expiration: %ld",
 		     TLScontext->namaddr, (long) key->tout);
     } else {
-	EVP_DecryptInit_ex(ctx, tkt_cipher, NOENGINE, key->bits, iv);
+	EVP_DecryptInit_ex(ctx, ciph, NOENGINE, key->bits, iv);
 	if (TLScontext->log_mask & TLS_LOG_CACHE)
 	    msg_info("%s: Decrypting session ticket, key expiration: %ld",
 		     TLScontext->namaddr, (long) key->tout);
@@ -347,8 +349,6 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
     int     scache_timeout;
     int     ticketable = 0;
     int     protomask;
-    int     min_proto;
-    int     max_proto;
     TLS_APPL_STATE *app_ctx;
     int     log_mask;
 
@@ -370,17 +370,21 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
      */
     tls_check_version();
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+
     /*
-     * Initialize the OpenSSL library, possibly loading its configuration
-     * file.
+     * Initialize the OpenSSL library by the book! To start with, we must
+     * initialize the algorithms. We want cleartext error messages instead of
+     * just error codes, so we load the error_strings.
      */
-    if (tls_library_init() == 0)
-	return (0);
+    SSL_load_error_strings();
+    OpenSSL_add_ssl_algorithms();
+#endif
 
     /*
      * First validate the protocols. If these are invalid, we can't continue.
      */
-    protomask = tls_proto_mask_lims(props->protocols, &min_proto, &max_proto);
+    protomask = tls_protocol_mask(props->protocols);
     if (protomask == TLS_PROTOCOL_INVALID) {
 	/* tls_protocol_mask() logs no warning. */
 	msg_warn("Invalid TLS protocol list \"%s\": disabling TLS support",
@@ -473,15 +477,6 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
     if (scache_timeout <= 0)
 	cachable = 0;
 
-    /*
-     * Presently we use TLS only with SMTP where truncation attacks are not
-     * possible as a result of application framing.  If we ever use TLS in
-     * some other application protocol where truncation could be relevant,
-     * we'd need to disable truncation detection conditionally, or explicitly
-     * clear the option in that code path.
-     */
-    off |= SSL_OP_IGNORE_UNEXPECTED_EOF;
-
     /*
      * Protocol work-arounds, OpenSSL version dependent.
      */
@@ -491,20 +486,18 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
      * Add SSL_OP_NO_TICKET when the timeout is zero or library support is
      * incomplete.
      */
+#ifdef SSL_OP_NO_TICKET
 #ifndef OPENSSL_NO_TLSEXT
     ticketable = (*var_tls_tkt_cipher && scache_timeout > 0
 		  && !(off & SSL_OP_NO_TICKET));
     if (ticketable) {
-#if OPENSSL_VERSION_PREREQ(3,0)
-	tkt_cipher = EVP_CIPHER_fetch(NULL, var_tls_tkt_cipher, NULL);
-#else
-	tkt_cipher = EVP_get_cipherbyname(var_tls_tkt_cipher);
-#endif
-	if (tkt_cipher == 0
-	    || EVP_CIPHER_mode(tkt_cipher) != EVP_CIPH_CBC_MODE
-	    || EVP_CIPHER_iv_length(tkt_cipher) != TLS_TICKET_IVLEN
-	    || EVP_CIPHER_key_length(tkt_cipher) < TLS_TICKET_IVLEN
-	    || EVP_CIPHER_key_length(tkt_cipher) > TLS_TICKET_KEYLEN) {
+	const EVP_CIPHER *ciph;
+
+	if ((ciph = EVP_get_cipherbyname(var_tls_tkt_cipher)) == 0
+	    || EVP_CIPHER_mode(ciph) != EVP_CIPH_CBC_MODE
+	    || EVP_CIPHER_iv_length(ciph) != TLS_TICKET_IVLEN
+	    || EVP_CIPHER_key_length(ciph) < TLS_TICKET_IVLEN
+	    || EVP_CIPHER_key_length(ciph) > TLS_TICKET_KEYLEN) {
 	    msg_warn("%s: invalid value: %s; session tickets disabled",
 		     VAR_TLS_TKT_CIPHER, var_tls_tkt_cipher);
 	    ticketable = 0;
@@ -530,6 +523,7 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
 #endif
     if (!ticketable)
 	off |= SSL_OP_NO_TICKET;
+#endif
 
     SSL_CTX_set_options(server_ctx, off);
 
@@ -538,10 +532,6 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
      */
     if (protomask != 0)
 	SSL_CTX_set_options(server_ctx, TLS_SSL_OP_PROTOMASK(protomask));
-    SSL_CTX_set_min_proto_version(server_ctx, min_proto);
-    SSL_CTX_set_max_proto_version(server_ctx, max_proto);
-    SSL_CTX_set_min_proto_version(sni_ctx, min_proto);
-    SSL_CTX_set_max_proto_version(sni_ctx, max_proto);
 
     /*
      * Some sites may want to give the client less rope. On the other hand,
@@ -617,6 +607,20 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
 	return (0);
     }
 
+    /*
+     * 2015-12-05: Ephemeral RSA removed from OpenSSL 1.1.0-dev
+     */
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+
+    /*
+     * According to OpenSSL documentation, a temporary RSA key is needed when
+     * export ciphers are in use, because the certified key cannot be
+     * directly used.
+     */
+    SSL_CTX_set_tmp_rsa_callback(server_ctx, tls_tmp_rsa_cb);
+    SSL_CTX_set_tmp_rsa_callback(sni_ctx, tls_tmp_rsa_cb);
+#endif
+
     /*
      * Diffie-Hellman key generation parameters can either be loaded from
      * files (preferred) or taken from compiled in values. First, set the
@@ -625,17 +629,19 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
      * the error handling, since we do have default values compiled in, so we
      * will not abort but just log the error message.
      */
+    SSL_CTX_set_tmp_dh_callback(server_ctx, tls_tmp_dh_cb);
+    SSL_CTX_set_tmp_dh_callback(sni_ctx, tls_tmp_dh_cb);
     if (*props->dh1024_param_file != 0)
-	tls_set_dh_from_file(props->dh1024_param_file);
-    tls_tmp_dh(server_ctx);
-    tls_tmp_dh(sni_ctx);
+	tls_set_dh_from_file(props->dh1024_param_file, 1024);
+    if (*props->dh512_param_file != 0)
+	tls_set_dh_from_file(props->dh512_param_file, 512);
 
     /*
      * Enable EECDH if available, errors are not fatal, we just keep going
      * with any remaining key-exchange algorithms.
      */
-    tls_auto_eecdh_curves(server_ctx, var_tls_eecdh_auto);
-    tls_auto_eecdh_curves(sni_ctx, var_tls_eecdh_auto);
+    tls_set_eecdh_curve(server_ctx, props->eecdh_grade);
+    tls_set_eecdh_curve(sni_ctx, props->eecdh_grade);
 
     /*
      * If we want to check client certificates, we have to indicate it in
@@ -718,7 +724,6 @@ TLS_APPL_STATE *tls_server_init(const TLS_SERVER_INIT_PROPS *props)
 				       sizeof(server_session_id_context));
 	SSL_CTX_set_session_cache_mode(server_ctx,
 				       SSL_SESS_CACHE_SERVER |
-				       SSL_SESS_CACHE_NO_INTERNAL |
 				       SSL_SESS_CACHE_NO_AUTO_CLEAR);
 	if (cachable) {
 	    app_ctx->cache_type = mystrdup(props->cache_type);
